#include <stdio.h>
#include <stdlib.h>
#include <cuda_runtime.h>
#include <cublas_v2.h>

#define REPEAT	1
#define SIZE	32768

float *matrix_alloc(void)
{
	float *a = malloc(SIZE * SIZE * sizeof(float));

	if (a == NULL)
		perror("matrix_alloc()");

	return a;
}

void matrix_random(float *a)
{
	static unsigned long r = 0x92a834c7df2e983d;

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			r = (r * 25214903917 + 11) & 0xffffffffffff;
			a[i * SIZE + j] = (float)(r >> 16) / 0xffffffff;
		}
	}
}

float matrix_sum(const float *a)
{
	float sum = 0.0;

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			sum = sum + a[i * SIZE + j];
		}
	}

	return sum;
}

cublasHandle_t handle;

int matrix_multiply7(float *restrict x,
	const float *restrict a, const float * restrict b)
{
	float *ga, *gb, *gx;

	cublasStatus_t sa, sb, sx;
	
	sa = cudaMalloc((void **)&ga, SIZE * SIZE * sizeof(float));
	sb = cudaMalloc((void **)&gb, SIZE * SIZE * sizeof(float));
	sx = cudaMalloc((void **)&gx, SIZE * SIZE * sizeof(float));

	if ((sa != cudaSuccess) || (sb != sa) || (sx != sa)) {
		fprintf(stderr, "cublas alloc failed\n");
		return -1;
	}

	// cuBLAS is column-major -- we ignore this and compute
	// everything transposed
	sa = cublasSetMatrix(SIZE, SIZE, sizeof(float), a, SIZE, ga, SIZE);
	sb = cublasSetMatrix(SIZE, SIZE, sizeof(float), b, SIZE, gb, SIZE);

	if ((sa != cudaSuccess) || (sb != sa)) {
		fprintf(stderr, "cublas download failed\n");
		return -1;
	}

	float alpha = 1.0;
	float beta = 0.0;

	sx = cublasSgemm(handle, CUBLAS_OP_N, CUBLAS_OP_N,
		SIZE, SIZE, SIZE, &alpha, ga, SIZE, gb, SIZE,
		&beta, gx, SIZE);

	if (sx != cudaSuccess) {
		fprintf(stderr, "cublas dgemm failed\n");
		return -1;
	}


	sx = cublasGetMatrix(SIZE, SIZE, sizeof(float), gx, SIZE, x, SIZE);

	if (sx != cudaSuccess) {
		fprintf(stderr, "cublas upload failed\n");
		return -1;
	}

	return 0;
}


int main(void)
{
	if (cublasCreate(&handle) != CUBLAS_STATUS_SUCCESS) {
		fprintf(stderr, "cublas init failed\n");
		return 1;
	}

	float *a = matrix_alloc();
	float *b = matrix_alloc();
	float *x = matrix_alloc();

	matrix_random(a);
	matrix_random(b);

	for (int r = 0; r < REPEAT; r++) {
		if (matrix_multiply7(x, a, b))
			return 1;
	}

	printf("%g\n", matrix_sum(x));

	free(a);
	free(b);
	free(x);

	cublasDestroy(handle);

	return 0;
}

