#define _POSIX_C_SOURCE 20231114L
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#define REPEAT	1
#define SIZE	1024

uint64_t now(void)
{
	struct timespec t;

	clock_gettime(CLOCK_REALTIME, &t);

	return (uint64_t)t.tv_nsec + (uint64_t)t.tv_sec * 1000000000;
}


double *matrix_alloc(void)
{
	double *a = malloc(SIZE * SIZE * sizeof(double));

	if (a == NULL)
		perror("matrix_alloc()");

	return a;
}

void matrix_random(double *a)
{
	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			a[i * SIZE + j] = (double)rand() / RAND_MAX;
		}
	}
}

double matrix_sum(double *a)
{
	double sum = 0.0;

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			sum = sum + a[i * SIZE + j];
		}
	}

	return sum;
}

void matrix_multiply0(double *x, double *a, double *b)
{
	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			double sum = 0.0;

			for (int k = 0; k < SIZE; k++) {
				sum += a[i * SIZE + k] * b[k * SIZE + j];
			}

			x[i * SIZE + j] = sum;
		}
	}
}

void matrix_multiply1(double *x, double *a, double *b)
{
	for (int i = 0; i < SIZE; i++) {
		for (int j = i + 1; j < SIZE; j++) {
			double tmp = b[i * SIZE + j];
			b[i * SIZE + j] = b[j * SIZE + i];
			b[j * SIZE + i] = tmp;
		}
	}

	for (int i = 0; i < SIZE; i++) {
		for (int j = 0; j < SIZE; j++) {
			double sum = 0.0;

			for (int k = 0; k < SIZE; k++) {
				sum += a[i * SIZE + k] * b[j * SIZE + k];
			}

			x[i * SIZE + j] = sum;
		}
	}
}

#define BLOCK	8
#define GRID	(SIZE / BLOCK)

static void block_zero(double *wx)
{
	for (int i = 0; i < BLOCK; i++) {
		for (int j = 0; j < BLOCK; j++) {
			wx[i * BLOCK + j] = 0.0;
		}
	}
}

static void block_fma(double *restrict wx,
	double *restrict wa, double *restrict wb)
{
	for (int i = 0; i < BLOCK; i++) {
		for (int j = 0; j < BLOCK; j++) {
			double sum = 0.0;

			for (int k = 0; k < BLOCK; k++) {
				sum += wa[i * BLOCK + k] * wb[j * BLOCK + k];
			}

			wx[i * BLOCK + j] += sum;
		}
	}
}

void matrix_multiply2(double *restrict x,
	double *restrict a, double * restrict b)
{
	double *wa = matrix_alloc();
	double *wb = matrix_alloc();
	double *wx = matrix_alloc();

	for (int wi = 0; wi < GRID; wi++) {
		for (int wj = 0; wj < GRID; wj++) {
			int offset = (wi * GRID + wj) * BLOCK * BLOCK;

			for (int li = 0; li < BLOCK; li++) {
				for (int lj = 0; lj < BLOCK; lj++) {
					int i = wi * BLOCK + li;
					int j = wj * BLOCK + lj;

					wa[offset + li * BLOCK + lj]
						= a[i * SIZE + j];
				}
			}
		}
	}

	for (int wj = 0; wj < GRID; wj++) {
		for (int wi = 0; wi < GRID; wi++) {
			int offset = (wj * GRID + wi) * BLOCK * BLOCK;

			for (int lj = 0; lj < BLOCK; lj++) {
				for (int li = 0; li < BLOCK; li++) {
					int i = wi * BLOCK + li;
					int j = wj * BLOCK + lj;

					wb[offset + lj * BLOCK + li]
						= b[i * SIZE + j];
				}
			}
		}
	}

	for (int wi = 0; wi < GRID; wi++) {
		for (int wj = 0; wj < GRID; wj++) {
			int xo = (wi * GRID + wj) * BLOCK * BLOCK;

			block_zero(wx + xo);

			for (int wk = 0; wk < GRID; wk++) {
				int ao = (wi * GRID + wk) * BLOCK * BLOCK;
				int bo = (wj * GRID + wk) * BLOCK * BLOCK;

				block_fma(wx + xo, wa + ao, wb + bo);
			}
		}
	}

	for (int wi = 0; wi < GRID; wi++) {
		for (int wj = 0; wj < GRID; wj++) {
			int id = wi * GRID + wj;
			int offset = id * BLOCK * BLOCK;

			for (int li = 0; li < BLOCK; li++) {
				for (int lj = 0; lj < BLOCK; lj++) {
					int i = wi * BLOCK + li;
					int j = wj * BLOCK + lj;

					x[i * SIZE + j] =
						wx[offset + li * BLOCK + lj];
				}
			}
		}
	}

	free(wa);
	free(wb);
	free(wx);
}

int main(void)
{
	double *a = matrix_alloc();
	double *b = matrix_alloc();
	double *x = matrix_alloc();

	matrix_random(a);
	matrix_random(b);

	for (int r = 0; r < REPEAT; r++) {
		matrix_multiply0(x, a, b);
	}

	printf("%g\n", matrix_sum(x));

	free(a);
	free(b);
	free(x);

	return 0;
}
